from numpy import *
import numpy.linalg as la

def check_unitarity(u,v):   #sanity check for u,v matrices
  N=len(u[:,0])
  unit=dot(conjugate(v),transpose(v))+dot(u,conjugate(transpose(u)))
  zero=dot(transpose(u),v) + dot(transpose(v),u)
  for i in range(N):
    for j in range(N):
     if i == j :
       if abs(unit[i,i]-1) > 0.0001 :
         print N,i,unit[i,i]
         raise Exception
     else :
       if abs(unit[i,j]) > 0.0001 :
         print shape(u),i,j,unit[i,j]
         raise Exception
     if abs(zero[i,j]) > 0.0001:
         print i,j,zero[i,j]
         raise Exception
         
def read_uv(N,filename):     # read in initial u,v, set
  lines=open(filename).readlines() 
  idx=0
  u=zeros(shape=(N,N) )
  v=zeros(shape=(N,N))
  Np=int(lines[idx][:5])
  if N != Np : raise Exception  
  idx +=1
  for i in range(N):
      ss=lines[idx].split()
      idx += 1
      for j in range(N):
        u[i,j] = float(ss[j])
  for i in range(N):
      ss=lines[idx].split()
      idx += 1
      for j in range(N):
        v[i,j] = float(ss[j])
  check_unitarity(u,v)
  return u,v  

def read_zc(N,filename):     
# read in  zc as z_r first, z_i next
  lines=open(filename).readlines() 
  zc=array([[0.0j]*N]*N )
  idx = 0
  if len(lines) != 2*N :
    print N,len(lines)
    raise Exception
  for i in range(N):
      ss=lines[idx].split()
      idx += 1
      for j in range(N):
        zc[i,j] = float(ss[j])
  for i in range(N):
      ss=lines[idx].split()
      idx += 1
      for j in range(N):
        zc[i,j] += float(ss[j])*1.0j
  return zc  

def Zfull(Z,uin,vin) :   #  Implements Eq. (27)
  u1=uin+ dot(vin,Z)
  v1=vin + dot(uin,Z)
  N=shape(Z)[0]
  unit = identity(N)
  onezz=unit+dot(transpose(Z),Z)
  e,vec=la.eigh(onezz)   
  emat=Z*0.0  
  for i in range(N) :
    emat[i,i] = e[i]**(-0.5)
  t1=dot(vec,emat)
  zu=dot(t1,transpose(vec))
  u2=dot(u1,zu); v2=dot(v1,zu)
  return u2,v2

def Zc(Z,U,V) :   
  N=shape(Z)[0]
  unit = identity(N)
  onezz=unit-dot(Z,conjugate(Z))
  e,vec=la.eigh(onezz)
  emat=Z*0.0  
  for i in range(N) :
    emat[i,i] = e[i]**(-0.5)
  t1=dot(vec,emat)
  zu=dot(t1,transpose(vec))
  up= U + conjugate(dot(V,Z))
  vp= V + conjugate(dot(U,Z))
  unew = dot(up,zu)
  vnew = dot(vp,zu)
  Zcal=array([[0.0j]*(2*N)]*(2*N))
  Zcal[:N,:N] += zu
  Zcal[N:,N:] += conjugate(zu)
  zzu=dot(Z,conjugate(zu))
  Zcal[:N,N:] = zzu
  Zcal[N:,:N] = conjugate(zzu)
  return unew,vnew

def pnice(mat,title='pnice') :     #print out matrix, nicely formated
  dim=shape(mat)[0]
  print title 
  for i in range(dim) :
    ss=''
    for j in range(dim):
       t1=mat[i,j] 
       if abs(t1) < 0.0001 :
         ss +='   0   ' 
       else :
         ss +=' %8.5f' % mat[i,j]
    print ss
  print
  
def rho_kappa(u,v) :   # Implements Eq. (3,4) of hfb_formulas.5.ps
  rho=dot(conjugate(v),transpose(v) )
  kappa=dot(conjugate(v),transpose(u) )
  return rho,kappa

def adjoint(U) :
  return conjugate(transpose(U))

def H20(u,v,hsp,delta) :    #Implements Eq. (22)
  h20=dot(adjoint(u),dot(hsp,conjugate(v)) )
  h20 += -dot(adjoint(v),dot(transpose(hsp),conjugate(u)) )
  h20 += dot(adjoint(u),dot(delta,conjugate(u)) )
  h20 += -dot(adjoint(v),dot(conjugate(delta),conjugate(v)) )
  return h20

def F20(u,v,F):
  f20 = dot(adjoint(u),dot(F,conjugate(v)) )
  f20 += -dot(adjoint(v),dot(F,conjugate(u)) )
  return f20

def F00(rho,Fop) :    # Implements Eq. (7)
  Fexp = trace(dot(Fop,rho))
  return real(Fexp)

def F11(u,v,Fop) :  # 
    uadj=transpose(u)
    vadj=transpose(v)
    t1=dot(transpose(Fop),v)
    t2=dot(vadj,t1)
    t3=dot(Fop,u)
    t4=dot(uadj,t3)
    f11=t4-t2
    return f11

def H00(rr,kk,vrr,dd,espl) :
  esp_rho=0.0
  v_rho=0.0
  vpair=0.0
  v_rho += 0.5*trace(dot(vrr,rr) )
  esp_rho += trace(dot(espl,rr) )
  vpair += -0.5*trace(dot(dd,conjugate(kk) ) )
  return esp_rho,v_rho,vpair
   
def write_uv(u,v,filename='uv.out'):  #final u,v set is written to uv.out
  f1=open(filename,'w')
  N=shape(u)[0]
  f1.write(' %i\n' % N)
  for i1 in range(N) :
      ss=''
      for i2 in range(N) :
        ss +=' %10.7f' % u[i1,i2]
      f1.write(ss+'\n')
  for i1 in range(N) :
      ss=''
      for i2 in range(N) :
        ss +=' %10.7f' % v[i1,i2]
      f1.write(ss+'\n')

def G11(u,v,Gop) :   #
    uadj=transpose(u)
    vadj=transpose(v)
    t1=dot(Gop,u)
    t2=dot(vadj,t1)
    t3=dot(Gop,v)
    t4=dot(uadj,t3)
    g11=t4-t2
    return g11



def H11(u,v,gamma,delta):
  h11= F11(u,v,gamma)
  temp = G11(u,v,delta)
  return h11 + temp
